# -*- mode: python; coding: utf-8 -*-
# Pigment rendering library
#
# Copyright © 2006 Fluendo Embedded S.L. <www.fluendo.com>
# All rights reserved.
#
# This file is available under one of two license agreements.
#
# This file is licensed under the GPL version 2. See "LICENSE.GPL" in
# the root of this distribution including a special exception to use
# Pigment with Fluendo's plugins.
#
# The GPL part of Pigment is also available under a commercial licensing
# agreement from Fluendo. See "LICENSE.Pigment" in the root directory
# of this distribution package for details on that license.

"""
Handle RGBA colors.
"""

from pgm.util import maths
from pgm.form import paint


# Indexation constants
R, G, B, A = 0, 1, 2, 3

# Predefined colors
BLACK = (0, 0, 0, 255)
RED = (255, 0, 0, 255)
GREEN = (0, 255, 0, 255)
BLUE = (0, 0, 255, 255)
GRAY = (128, 128, 128, 255)
CYAN = (0, 255, 255, 255)
MAGENTA = (255, 0, 255, 255)
YELLOW = (255, 255, 0, 255)
WHITE = (255, 255, 255, 255)


class Color(paint.Paint):
    """
    """

    def __init__(self, color=WHITE):
        paint.Paint.__init__(self, opacity=color[A])
        self._r = maths.clamp(int(color[R]), 0, 255)
        self._g = maths.clamp(int(color[G]), 0, 255)
        self._b = maths.clamp(int(color[B]), 0, 255)
        self._a = maths.clamp(int(color[A]), 0, 255)
        self._set_surface_color()

    def __repr__(self):
        return "<color.Color((%s, %s, %s, %s))>" \
               % (self._r, self._g, self._b, self._a)

    # Private methods

    def _set_surface_color(self):
        for shape in self._shape:
            shape._surface.set_media_color(self._r, self._g, self._b, self._a)

    # Property definitions

    def r__get(self):
        """The R color layer"""
        return self._r

    def r__set(self, r):
        self._r = maths.clamp(int(r), 0, 255)
        self._set_surface_color()

    def g__get(self):
        """The G color layer"""
        return self._g

    def g__set(self, g):
        self._g = maths.clamp(int(g), 0, 255)
        self._set_surface_color()

    def b__get(self):
        """The B color layer"""
        return self._b

    def b__set(self, b):
        self._b = maths.clamp(int(b), 0, 255)
        self._set_surface_color()

    def a__get(self):
        """The A color layer"""
        return self._a

    def a__set(self, a):
        self._a = maths.clamp(int(a), 0, 255)
        self._opacity = self._a
        self._set_surface_color()

    def color__get(self):
        """The color"""
        return (self._r, self._g, self._b, self._a)

    def color__set(self, color):
        self._r = maths.clamp(int(color[R]), 0, 255)
        self._g = maths.clamp(int(color[G]), 0, 255)
        self._b = maths.clamp(int(color[B]), 0, 255)
        self._a = maths.clamp(int(color[A]), 0, 255)
        self._set_surface_color()

    def opacity__set(self, opacity):
        paint.Paint.opacity__set(self, opacity)
        self._a = int(self._opacity)
        self._set_surface_color()
        
# Color utility functions

def add(color1, color2):
    """
    """
    return Color((maths.clamp(int(color1.r + color2.r), 0, 255),
                  maths.clamp(int(color1.g + color2.g), 0, 255),
                  maths.clamp(int(color1.b + color2.b), 0, 255),
                  maths.clamp(int(color1.a + color2.a), 0, 255)))

def ponderate(color, factor):
    """
    """
    return Color((maths.clamp(int(color.r * factor), 0, 255),
                  maths.clamp(int(color.g * factor), 0, 255),
                  maths.clamp(int(color.b * factor), 0, 255),
                  maths.clamp(int(color.a * factor), 0, 255)))


# module testing

if __name__ == '__main__':
    print ponderate(Color(RED), 0.5)
    print ponderate(Color((100, 50, 25, 255)), 2)
    print add(Color(RED), Color(BLUE))
