# -*- mode: python; coding: utf-8 -*-
# Pigment rendering library
#
# Copyright © 2006 Fluendo Embedded S.L. <www.fluendo.com>
# All rights reserved.
#
# This file is available under one of two license agreements.
#
# This file is licensed under the GPL version 2. See "LICENSE.GPL" in
# the root of this distribution including a special exception to use
# Pigment with Fluendo's plugins.
#
# The GPL part of Pigment is also available under a commercial licensing
# agreement from Fluendo. See "LICENSE.Pigment" in the root directory
# of this distribution package for details on that license.

"""
"""

from pgm import widget, log, render
from pgm.util import maths
from pgm.layout import leaf, window
from pgm.form import color, stroke, shadow, image, gradient, stream
import pgmrender
import threading


class Shape(leaf.Leaf):
    """
    """

    def __init__(self, size, position, paint, stroke, shadow, media):
        self._surface = render.get_renderer().create_surface()
	leaf.Leaf.__init__(self, size, position)
	# Create the shape parameters
	self._paint = None
	self._stroke = None
	self._shadow = None
	self._media = None
	self._media_lock = threading.Lock()
	# Initialize the shape parameters
        self.paint = paint
        self.stroke = stroke
	self.shadow = shadow
        self.media = media
        self.aspect_style = pgmrender.LAYOUT_SCALED

    # Property definitions

    def paint__get(self):
	"""The paint of the shape"""
	return self._paint

    def paint__set(self, paint):
	if paint == self._paint:
	    log.warn('Shape.paint__set(): %r is the actual paint' % paint)
	    return
	if isinstance(paint, (color.Color, gradient.Gradient)):
	    # Apply the color to the surface
	    self._surface.set_media_color(paint.r, paint.g, paint.b, paint.a)
	    paint._add_shape(self)
	    self._paint = paint
	else:
	    log.warn('Shape.paint__set(): paint must be Color or Gradient')
	    self._paint = None

    def stroke__get(self):
	"""The stroke of the shape"""
        return self._stroke

    def stroke__set(self, _stroke):
	if _stroke == self._stroke:
	    log.warn('Shape.stroke__set(): %r is the actual stroke' % _stroke)
	    return
	if isinstance(_stroke, stroke.Stroke):
	    # Apply the stroke to the surface
            log.dbg("%r, %r", self._surface, _stroke.width)
	    self._surface.set_width(_stroke.width)
	    self._surface.set_stroke_color(_stroke.color.r, _stroke.color.g,
					   _stroke.color.b, _stroke.color.a)
	    self._surface.set_pattern(_stroke.pattern)
	    self._surface.set_factor(_stroke.factor)
	    _stroke._add_shape(self)
	    self._stroke = _stroke
	else:
	    log.warn('Shape.stroke__set(): stroke must be Stroke')
	    self._paint = None

    def media__get(self):
        """
	The media of the shape
	media is thread safe	
	"""
        media = None
        try:
            self._media_lock.acquire()
            media = self._media
        finally:
            self._media_lock.release()
            return media

    def media__set(self, media):
	"""media is threadsafe"""
        if media == self._media:
            return
	self._media_lock.acquire()
	if isinstance(media, image.Image):
	    if self._media != None:
		self._surface.del_image()
                self._media._del_shape(self)
	    media._add_shape(self)
	    self._media = media
            if len(media._shape) <= 1:
                if media.pixels:
                    self._surface.set_image(media.format, media.size[0],
                                            media.size[1], 0, media.pixels)
            else:
                self._surface.set_image_from_drawable(media._shape[0]._surface)
        elif isinstance(media, stream.Stream):
            if self._media != None:
                self._surface.del_image()
                self._media._del_shape(self)
            media._add_shape(self)
            self._media = media
            if len(media._shape) <= 1:
                self._media.drawable = self._surface
            else:
                self._surface.set_image_from_drawable(media._shape[0]._surface)
	else:
	    log.warn('Shape.media__set(): media must be Image or Stream')
	    self._media = None
	self._media_lock.release()

    def shadow__get(self):
        """The shadow of the shape"""
	return self._shadow

    def shadow__set(self, _shadow):
	if isinstance(_shadow, shadow.Shadow):
	    _shadow._add_shape(self)
	    self._shadow = _shadow
	else:
	    log.warn('Shape.shadow__set(): shadow must be Shadow')
	    self._shadow = None

    def aspect_style__get(self):
        """The aspect style of the binded media"""
        return self._aspect_style

    def aspect_style__set(self, style):
        if (style == pgmrender.LAYOUT_SCALED
            or style == pgmrender.LAYOUT_FILLED
            or style == pgmrender.LAYOUT_CENTERED):
            self._aspect_style = style
            self._surface.set_aspect_style(style)

    def visible__set(self, visible):
        widget.Widget.visible__set(self, visible)
        if visible:
            self._surface.show()
        else:
            self._surface.hide()

    # Property overriding

    def width__set(self, width):
        widget.Widget.width__set(self, width)
        self._surface.set_size(*self.size)

    def height__set(self, height):
        widget.Widget.height__set(self, height)
        self._surface.set_size(*self.size)

    def size__set(self, size):
        relative_width = size[0] - self.width__get()
        relative_height = size[1] - self.height__get()
        widget.Widget.size__set(self, size)
        size_mode = widget.Widget.size_mode__get(self)
        if size_mode == widget.BOTTOM_RIGHT:
            pass
        if size_mode == widget.BOTTOM_LEFT:
            self.x -= relative_width
        elif size_mode == widget.UP_LEFT:
            self.x -= relative_width
            self.y -= relative_height
        elif size_mode == widget.UP_RIGHT:
            self.y -= relative_width
        elif size_mode == widget.CENTER:
            self.x -= relative_width / 2.0
            self.y -= relative_height / 2.0
        elif size_mode == widget.UP_CENTER:
            self.x -= relative_width / 2.0
        elif size_mode == widget.BOTTOM_CENTER:
            self.x -= relative_width / 2.0
            self.y -= relative_height
        elif size_mode == widget.LEFT_CENTER:
            self.y -= relative_height / 2.0
        elif size_mode == widget.RIGHT_CENTER:
            self.x -= relative_width 
            self.y -= relative_height / 2.0
        self._surface.set_size(*self.size)

    def x__set(self, x):
        widget.Widget.x__set(self, x)
        self._surface.set_position(*self.position)

    def y__set(self, y):
        widget.Widget.y__set(self, y)
        self._surface.set_position(*self.position)

    def position__set(self, position):
        widget.Widget.position__set(self, position)
        self._surface.set_position(*self.position)

    def tx__set(self, tx):
        widget.Widget.tx__set(self, tx)
        self._surface.set_translation(*self.translation)

    def ty__set(self, ty):
        widget.Widget.ty__set(self, ty)
        self._surface.set_translation(*self.translation)

    def tz__set(self, tz):
        widget.Widget.tz__set(self, tz)
        self._surface.set_translation(*self.translation)

    def translation__set(self, translation):
        widget.Widget.translation__set(self, translation)
        self._surface.set_translation(*self.translation)

    def rx__set(self, rx):
        widget.Widget.rx__set(self, rx)
        self._surface.set_rotation(*self.rotation)

    def ry__set(self, ry):
        widget.Widget.ry__set(self, ry)
        self._surface.set_rotation(*self.rotation)

    def rz__set(self, rz):
        widget.Widget.rz__set(self, rz)
        self._surface.set_rotation(*self.rotation)

    def rotation__set(self, rotation):
        widget.Widget.rotation__set(self, rotation)
        self._surface.set_rotation(*self.rotation)

    def opacity__set(self, opacity):
        widget.Widget.opacity__set(self, opacity)
        self._paint.opacity = self.opacity__get()
        
    def opacity_offset__set(self, opacity_offset):
        widget.Widget.opacity_offset__set(self, opacity_offset)
        self._paint.opacity = self.opacity__get()

    def x_align__set(self, x_align):
	leaf.Leaf.x_align__set(self, x_align)
	self._surface.set_alignment(self._x_align, self._y_align)

    def y_align__set(self, y_align):
	leaf.Leaf.y_align__set(self, y_align)
	self._surface.set_alignment(self._x_align, self._y_align)

    def align__set(self, align):
	leaf.Leaf.align__set(self, align)
	self._surface.set_alignment(self._x_align, self._y_align)

    def ordering__set(self, ordering):
        widget.Widget.ordering__set(self, ordering)
        if self._window:
            self._window._rdr_window.del_drawable(self._surface)
            self._window._rdr_window.add_drawable(self._surface,
                                                  widget.order[self._ordering])
            
    # Public methods overriding

    def ponderate_opacity(self, ponderation):
        """
        """
        widget.Widget.ponderate_opacity(self, ponderation)
        self._paint.opacity *= ponderation

    # Private methods

    def _set_window(self, window):
        self._window = window
        window._rdr_window.add_drawable(self._surface,
                                        widget.order[self._ordering])

    def _unset_window(self):
        if self._window:
            self._window._rdr_window.del_drawable(self._surface)
        self._window = None
