# -*- mode: python; coding: utf-8 -*-
# Pigment rendering library
#
# Copyright © 2006 Fluendo Embedded S.L. <www.fluendo.com>
# All rights reserved.
#
# This file is available under one of two license agreements.
#
# This file is licensed under the GPL version 2. See "LICENSE.GPL" in
# the root of this distribution including a special exception to use
# Pigment with Fluendo's plugins.
#
# The GPL part of Pigment is also available under a commercial licensing
# agreement from Fluendo. See "LICENSE.Pigment" in the root directory
# of this distribution package for details on that license.

"""
"""

from pgm.form import media, component
from pgm import log, render
from pgm.util import discoverer

# Check and import GStreamer
import pygst
pygst.require("0.10")
import gst
# Load the Pigment GStreamer video sink
gst.plugin_load_file('/usr/lib64/pigment-0.1/gstreamer/libpgmrendersink.so')


class Stream(media.Media):
    """
    """

    def __init__(self, location=None, pipeline=None, discovered_callback=None):
        media.Media.__init__(self, media.BILINEAR, 1.33)

        # setup internal variables
        self._drawable = None
        self._saved_volume = 0

        # configure the pipeline
        self.setup_pipeline(pipeline)

        # feed the pipeline
        if location:
            self.location = location

        # optional discoverer
        self.discovered_callback = discovered_callback
        if self.discovered_callback:
            self.discoverer = discoverer.Discoverer()
            self.discoverer.connect('discovered', self._discovered)
            
            try:
                self.discoverer.set_uri(location)
            except AssertionError:
                # FIXME: workaround for unsupported uris
                self.discoverer.is_audio = True
                self.discoverer.is_video = False
                self._discovered(None, None)
            else:
                self.discoverer.discover()

    def setup_pipeline(self, pipeline=None):
        if isinstance(pipeline, str):
            pipeline = gst.parse_launch(pipeline)
        elif not pipeline:
            # fallback to playbin
            pipeline = gst.element_factory_make('playbin', 'playbin')
            pgm_sink = gst.element_factory_make('pgmrendersink', 'sink')
            pipeline.set_property('video-sink', pgm_sink)
        self.pipeline = pipeline
        
    def _discovered(self, s, r):
        self.discovered_callback(self)

    # Property definitions
        
    def location__get(self):
        """The location of the audio/video stream"""
        return self._location

    def location__set(self, location):
        if isinstance(location, str):
            self._location = location
            self._src.set_property('uri', self._location)
        else:
            self._location = None

    def pipeline__set(self, pipeline):
        self._pipeline = pipeline
        src = self._pipeline.get_by_name('src')
        if not src:
            src = self._pipeline
        self._src = src
        
    def pipeline__get(self):
        """The GStreamer pipeline of the stream"""
        return self._pipeline

    def drawable__get(self):
        """The drawable to output the video stream"""
        return self._drawable

    def drawable__set(self, drawable):
        if self._pipeline:
            sink = self._pipeline.get_by_name('sink')
            if not sink:
                sink = self._pipeline.get_property('video-sink')
            if sink:
                sink.set_property('drawable', drawable)
            else:
                log.warn('Stream.drawable__set(): '
                         'No \"pgmrendersink\" in the GStreamer pipeline')

    def get_volume_element(self):
        vlm = None
        if self._pipeline:
            vlm = self._pipeline.get_by_name('vlm')
            if not vlm:
                # in case of playbin, volume element is wrapped inside
                vlm = self._pipeline
        return vlm

    def volume__get(self):
        """The volume of the audio stream"""
        volume = 0
        vlm = self.get_volume_element()
        if vlm:
            volume = vlm.get_property('volume')
        return volume

    def volume__set(self, volume):
        vlm = self.get_volume_element()
        if vlm:
            vlm.set_property('volume', volume)

    def mute__get(self):
        """The mute state of the audio stream"""
        muted = False
        vlm = self.get_volume_element()
        if vlm:
            # FIXME: couln't find gst_element.has_property() ...
            if 'mute' in dir(vlm.props):
                muted = vlm.get_property('mute')
            else:
                muted = self.volume == 0
        return muted

    def mute__set(self, muted):
        vlm = self.get_volume_element()
        if vlm:
            # FIXME: couln't find gst_element.has_property() ...
            if 'mute' in dir(vlm.props):
                vlm.set_property('mute', muted)
            else:
                if muted:                            
                    self._saved_volume = self.volume
                    self.volume = 0
                elif self._saved_volume:
                    self.volume = self._saved_volume
            return True
        return False
    
    # Public methods

    def play(self):
        if self._pipeline != None:
            self._pipeline.set_state(gst.STATE_PLAYING)

    def pause(self):
        if self._pipeline != None:
            self._pipeline.set_state(gst.STATE_PAUSED)

    def stop(self):
        if self._pipeline != None:
            self._pipeline.set_state(gst.STATE_READY)
