# -*- mode: python; coding: utf-8 -*-
# Pigment rendering library
#
# Copyright © 2006 Fluendo Embedded S.L. <www.fluendo.com>
# All rights reserved.
#
# This file is available under one of two license agreements.
#
# This file is licensed under the GPL version 2. See "LICENSE.GPL" in
# the root of this distribution including a special exception to use
# Pigment with Fluendo's plugins.
#
# The GPL part of Pigment is also available under a commercial licensing
# agreement from Fluendo. See "LICENSE.Pigment" in the root directory
# of this distribution package for details on that license.

"""
"""

from pgm import widget, log
from pgm.message import signal


class Group(widget.Widget):
    """
    """

    def __init__(self, size=(50,50), position=(0,0)):
        self._children = []
        self._x, self._y = 0, 0
        self._position = (self._x, self._y)
        widget.Widget.__init__(self, position=position, size=size)

        # Signal registration
        self.sig_add = signal.Signal('add', (widget.Widget,))
        self.sig_remove = signal.Signal('remove', (widget.Widget,))

    def __contains__(self, widget):
        return widget.parent == self

    # Property definitions

    def children__get(self):
        """The children list"""
        return self._children

    # Property overriding

    def x__set(self, x):
        offset = x - self.x
        widget.Widget.x__set(self, x)
        for child in self._children:
            child.x += offset

    def y__set(self, y):
        offset = y - self.y
        widget.Widget.y__set(self, y)
        for child in self._children:
            child.y += offset

    def position__set(self, position):
        offset = map(lambda x,y: x-y, position, self.position)
        widget.Widget.position__set(self, position)
        for child in self._children:
            child.position = map(lambda x,y: x+y, child.position, offset)

    def tx__set(self, tx):
        widget.Widget.tx__set(self, tx)
        for child in self._children:
            child.tx = tx

    def ty__set(self, ty):
        widget.Widget.ty__set(self, ty)
        for child in self._children:
            child.ty = ty

    def tz__set(self, tz):
        widget.Widget.tz__set(self, tz)
        for child in self._children:
            child.tz = tz

    def translation__set(self, translation):
        widget.Widget.translation__set(self, translation)
        for child in self._children:
            child.translation = translation

    def rx__set(self, rx):
        widget.Widget.rx__set(self, rx)
        for child in self._children:
            child.rx = rx

    def ry__set(self, ry):
        widget.Widget.ry__set(self, ry)
        for child in self._children:
            child.ry = ry

    def rz__set(self, rz):
        widget.Widget.rz__set(self, rz)
        for child in self._children:
            child.rz = rz

    def rotation__set(self, rotation):
        widget.Widget.rotation__set(self, rotation)
        for child in self._children:
            child.rotation = rotation

    def opacity__set(self, opacity):
        widget.Widget.opacity__set(self, opacity)
        for child in self._children:
            child.opacity = opacity

    def visible__set(self, visible):
        widget.Widget.visible__set(self, visible)
        for child in self._children:
            child.visible = visible

    def opacity_offset__set(self, opacity_offset):
        widget.Widget.opacity_offset__set(self, opacity_offset)
        for child in self._children:
            child.opacity_offset = opacity_offset
            
    def ordering__set(self, ordering):
        widget.Widget.ordering__set(self, ordering)
        if self._window:
            for child in self._children:
                child.ordering = ordering

    # Public methods

    def add(self, *args):
        """
        Add widget(s) to the group.
        """
        for widget in args:
            self._add(widget)

    def remove(self, widget):
        """
        Remove a widget from the group.
        """
        # Am I the parent of the widget ?
        if widget and widget.parent == self:
            self._children.remove(widget)
            widget._parent = None
            widget.position = (0, 0)
            # Already binded to a Window ? Let's remove the Widget.
            if self._window:
                widget._unset_window()
            log.dbg("Group.remove(): %s removed", widget)
            return True
        else:
            log.dbg("Group.remove(): %s not removed, not in %s", widget, self)
            return False

    def reorder_child(self, widget, position):
        """
        Reorder a widget of the group to the given position. If position is
        negative or superior to the size of the children list, the widget is
        will be reordered at the end of the list.
        """
        # Am I the parent of the widget ?
        if widget.parent == self:
            surface = self._children[self._children.index(widget)]._surface
            self._window._rdr_window.reorder_drawable(surface, position)

    # Public methods overriding

    def is_ancestor(self, widget):
        """
        FIXME
        """
        pass

    def ponderate_opacity(self, ponderation):
        """
        """
        widget.Widget.ponderate_opacity(ponderation)
        for child in self._children:
            child.ponderate_opacity(ponderation)

    # Private methods

    def _add(self, widget):
        # Already got a parent ?
        if widget.parent == None:
            # Add the group position to the widget one 
            widget.position = map(lambda x,y: x+y,
                                  widget.position, self.position)
            self._children.append(widget)
            widget._parent = self
            if self.visible == False:
                widget.visible = self.visible
            # Already binded to a Window ? Let's add the Widget.
            if self._window:
                widget._set_window(self._window)
            log.dbg("Group._add(): %s added", widget)
            return True
        else:
            log.dbg("Group._add(): %s not added", widget)
            return False

    def _set_window(self, window):
        """
        Bind a window to the group and its children.
        """
        self._window = window
        for child in self._children:
            child._set_window(window)

    def _unset_window(self):
        """
        Unbind the window of the group and its children.
        """
        self._window = None
        for child in self._children:
            child._unset_window()

    def _do_motion_event(self, event):
        """
        """
        for child in self._children:
            if child._widget._has_pointer(event):
                if child._widget._had_pointer:
                    child.sig_motion_notify_event.emit(event)
                else:
                    child.sig_enter_notify_event.emit(event)
                    child._widget._had_pointer = True
                widget.Widget._entered.append(child._widget)
                child._widget._do_motion_event(event)

    def _do_key_event(self, _event):
        """
        """
        # FIXME: Handle a focus system, ATM key event are forwarded to
        #        all the children widgets.
        signal_name = self.get_signal_name(_event)
        for child in self._children:
            signal = getattr(child._widget, signal_name)
            signal.emit(_event)
            child._widget._do_key_event(_event)

    def _do_button_event(self, _event):
        """
        """
        signal_name = self.get_signal_name(_event)
        for child in self._children:
            if child._widget._has_pointer(_event):
                signal = getattr(child._widget, signal_name)
                signal.emit(_event)
                child._widget._do_button_event(_event)

    def _do_scroll_event(self, event):
        """
        """
        for child in self._children:
            if child._widget._has_pointer(event):
                child._widget.sig_scroll_event.emit(event)
                child._widget._do_scroll_event(event)
