# -*- mode: python; coding: utf-8 -*-
# Pigment rendering library
#
# Copyright © 2006, 2007 Fluendo Embedded S.L. <www.fluendo.com>
# All rights reserved.
#
# This file is available under one of two license agreements.
#
# This file is licensed under the GPL version 2. See "LICENSE.GPL" in
# the root of this distribution including a special exception to use
# Pigment with Fluendo's plugins.
#
# The GPL part of Pigment is also available under a commercial licensing
# agreement from Fluendo. See "LICENSE.Pigment" in the root directory
# of this distribution package for details on that license.

"""
"""

from pgm import widget, render, log
from pgm.layout import group
from pgm.message import signal
from pgm.form import color
import pgmrender, gobject

# Scaling mode constants
ADJUST, CUSTOM, SCREEN_ADJUST = 0, 1, 2


class Window(group.Group):
    """
    """

    def __init__(self, size=None, title="Pigment Window",
                 fullscreen=False, color=color.Color(color.BLACK),
                 canvas_size=None, scale_mode=ADJUST):
        self._rdr_window = render.get_renderer().create_window()

        if not size:
            size = self._get_default_size()
        group.Group.__init__(self, size=size, position=(0,0))

        if not canvas_size:
            self.canvas_size = self.size
        self.scale_mode = scale_mode
        
        self.set_title(title, title)
        self.color = color
        self._window = self
        self._pushed_events = []
        self._mainloop = None
        self._is_pigment_loop = True

        # Signal definitions
        self.sig_resize_event = signal.Signal('resize-event', (object, tuple))
        self.sig_exit_event = signal.Signal('exit', (object,))

        self.set_fullscreen(fullscreen, self.size)

    # Property definitions

    def scale_mode__get(self):
        """FIXME"""
        scale_mode = self._rdr_window.get_scale_mode()
        if scale_mode == pgmrender.WINDOW_ADJUST:
            return ADJUST
        elif scale_mode == pgmrender.WINDOW_CUSTOM:
            return CUSTOM
        elif scale_mode == pgmrender.WINDOW_SCREEN_ADJUST:
            return SCREEN_ADJUST

    def scale_mode__set(self, scale_mode):
        """FIXME"""
        if scale_mode == ADJUST:
            self._rdr_window.set_scale_mode(pgmrender.WINDOW_ADJUST)
        elif scale_mode == CUSTOM:
            self._rdr_window.set_scale_mode(pgmrender.WINDOW_CUSTOM)
        elif scale_mode == SCREEN_ADJUST:
            self._rdr_window.set_scale_mode(2)

    def scale__get(self):
        """FIXME"""
        return self._rdr_window.get_scale()

    def scale__set(self, scale):
        """FIXME"""
        if self.scale_mode == CUSTOM:
            self._rdr_window.set_scale(scale[0], scale[1])

    def screen_size_mm__get(self):
        """FIXME"""
        return self._rdr_window.get_screen_size_mm()
    
    def screen_resolution__get(self):
        """FIXME"""
        return self._rdr_window.get_screen_resolution()

    def color__get(self):
        """The background clear color"""
        return self._color

    def color__set(self, _color):
        if isinstance(_color, color.Color):
            self._color = _color
            self._rdr_window.set_color(self._color.r, self._color.g,
                                       self._color.b, self._color.a)

    def cursor__get(self):
        """The cursor"""
        return self._rdr_window.get_cursor()

    def cursor__set(self, cursor):
        return self._rdr_window.set_cursor(cursor)

    def size__get(self):
        return self._rdr_window.get_size()

    def size__set(self, size):
        group.Group.size__set(self, size)
        self._rdr_window.set_size(size[0], size[1])

    def canvas_size__get(self):
        return self._rdr_window.get_canvas_size()

    def canvas_size__set(self, canvas_size):
        self._rdr_window.set_canvas_size(canvas_size[0], canvas_size[1])
        
    def screen_size__get(self):
        """The screen size in pixels"""
        return self._rdr_window.get_screen_size()

    def screen_size_mm__get(self):
        """The screen size in millimeters"""
        return self._rdr_window.get_screen_size_mm()

    def screen_resolution__get(self):
        """The screen size in pixels"""
        return self._rdr_window.get_screen_resolution()

    ### Public methods

    def regenerate_widgets(self):
        """
        """
        self._rdr_window.regenerate_drawables()

    def set_title(self, window_title, icon_title):
        """
        """
        self._window_title, self._icon_title = window_title, icon_title
        self._rdr_window.set_title(self._window_title, self._icon_title)

    def get_title(self):
        """
        """
        return (self._window_title, self._icon_title)

    def set_fullscreen(self, fullscreen, size=None):
        """
        """
        self._fullscreen = fullscreen
        if size == None:
            if fullscreen:
                size = self._window.screen_resolution
                self._tmp_window_size = self.size
            else:
                try:
                    size = self._tmp_window_size
                except AttributeError:
                    size = self._window._get_default_size()
        self._rdr_window.set_fullscreen(fullscreen, size[0], size[1])

    def get_fullscreen(self):
        """
        """
        return self._fullscreen

    def push_events(self, events):
        """
        Add an additional events list to be dispatched.
        """
        if isinstance(events, (list, tuple)):
            self._pushed_events += events

    def update(self):
        """
        Redraw the window.
        """
        self._rdr_window.update()
        return True

    def main(self, mainloop=None):
        """
        Run the mainloop until the main_quit method is called.
        If a mainloop is provided, we don't start the loop.
        """
        self._rdr_window.start_loop()
        # FIXME: The timeout period must be validated
        gobject.timeout_add(15, self._dispatch_events)
        if not mainloop:
            self._mainloop = gobject.MainLoop()
            self._mainloop.run()
        else:
            self._is_pigment_loop = False
            self._mainloop = mainloop

    def main_quit(self):
        """
        Terminate the mainloop.
        """
        self._rdr_window.quit_loop()
        if self._is_pigment_loop:
            self._mainloop.quit()

    ### Public inherited methods

    ### Private methods

    def _dispatch_events(self):
        events = self._pushed_events + self._rdr_window.get_events()
        for event in events:
            self._do_event(event)
        self._pushed_events = []
        return True

    def _get_default_size(self):
        screen_resolution = float(self.screen_resolution[0]), \
                            float(self.screen_resolution[1])
        if screen_resolution[0] <= 800:
            width = 640
        elif screen_resolution[0] <= 1024:
            width = 800
        else:
            width = 1024
        height = float(width) * (screen_resolution[1] / screen_resolution[0])
        return (width, int(height))

    def _do_event(self, _event):
        evt_type = _event.type
        if evt_type == pgmrender.MOTION_NOTIFY:
            self._do_motion_event(_event)
        elif evt_type == pgmrender.KEY_PRESS:
            self._do_key_press_event(_event)
        elif evt_type == pgmrender.KEY_RELEASE:
            self._do_key_release_event(_event)
        elif evt_type in (pgmrender.BUTTON_PRESS, pgmrender.BUTTON_RELEASE):
            self._do_button_event(_event)
        elif evt_type == pgmrender.SCROLL:
            self._do_scroll_event(_event)
        elif evt_type == pgmrender.CONFIGURE:
            self._do_configure_event(_event)
        elif evt_type == pgmrender.DELETE:
            self._do_exit_event(_event)
        else:
            print 'un-handled:', _event
            
    def _do_exit_event(self, event):
        self.sig_exit_event.emit(event)
            
    def _do_motion_event(self, event):
        self.sig_motion_notify_event.emit(event)
        for child in self._children:
            if child._has_pointer(event):
                if child._had_pointer:
                    child.sig_motion_notify_event.emit(event)
                else:
                    child.sig_enter_notify_event.emit(event)
                    child._had_pointer = True
                widget.Widget._entered.append(child)
                child._do_motion_event(event)
        for wdg in [wdg for wdg in widget.Widget._last_entered
                    if wdg not in widget.Widget._entered]:
            wdg._had_pointer = False
            wdg.sig_leave_notify_event.emit(event)
        widget.Widget._last_entered = widget.Widget._entered
        widget.Widget._entered = []

    def _do_key_press_event(self, _event):
        #FIXME: Handle a focus system
        signal_name = self.get_signal_name(_event)
        getattr(self, signal_name).emit(_event)        
##         for child in self._children:
##             print getattr(child, signal_name)
##             getattr(child, signal_name).emit(_event)
##             child._do_key_event(_event)

    def _do_key_release_event(self, _event):
        signal_name = self.get_signal_name(_event)
        getattr(self, signal_name).emit(_event)        

    def _do_button_event(self, _event):
        """
        TODO: Handle 2BUTTON_PRESS and 3BUTTON_PRESS too??
        """
        signal_name = self.get_signal_name(_event)
        getattr(self, signal_name).emit(_event)
        for child in self._children:
            if child._has_pointer(_event):
                getattr(child, signal_name).emit(_event)
                child._do_button_event(_event)

    def _do_scroll_event(self, event):
        self.sig_scroll_event.emit(event)
        for child in self._children:
            if child._has_pointer(event):
                child.sig_scroll_event.emit(event)
                child._do_scroll_event(event)

    def _do_configure_event(self, _event):
        self.sig_resize_event.emit(_event)
