# -*- mode: python; coding: utf-8 -*-
# Pigment rendering library
#
# Copyright © 2006 Fluendo Embedded S.L. <www.fluendo.com>
# All rights reserved.
#
# This file is available under one of two license agreements.
#
# This file is licensed under the GPL version 2. See "LICENSE.GPL" in
# the root of this distribution including a special exception to use
# Pigment with Fluendo's plugins.
#
# The GPL part of Pigment is also available under a commercial licensing
# agreement from Fluendo. See "LICENSE.Pigment" in the root directory
# of this distribution package for details on that license.

# TODO: - Spline interpolation

"""
"""

from pgm import log
from pgm.util import classinit, maths
from pgm.timing import keyframe

# Interpolation constants
LINEAR, SPLINE = 0, 1
ABSOLUTE, RELATIVE = 0, 1


class Modifier(object):
    """
    """

    # Allows property fget/fset/fdel/doc overriding
    __metaclass__ = classinit.ClassInitMeta
    __classinit__ = classinit.build_properties

    def __init__(self, obj=[], prop=None, timeline=[], interpolation=LINEAR, operation=ABSOLUTE):
        self._lerp = None
        self._operation = operation
        self._prop_initial_values = []
        
        self.obj = obj
        self.prop = prop
        self.timeline = timeline
        self.interpolation = interpolation

    def __repr__(self):
        string = {LINEAR:'LINEAR', SPLINE:'SPLINE'}
        return "<modifier.Modifier(obj=%r, prop=%r, timeline=%r, " \
               "interpolation=%s)>" % (self._obj, self._prop, self._timeline,
                                       string[self._interpolation])

    # Property definitions

    def obj__get(self):
        """The list of objects"""
        return self._obj

    def obj__set(self, obj):
        if isinstance(obj, (list, tuple)):
            try:
                obj_type = type(obj[0])
                for i in obj[1:]:
                    pass
##                     if (not issubclass(obj_type, type(i))) or (not issubclass(type(i),obj_type)):
##                         error = 'Mismatching types in the list (%r vs %r)'
##                         raise TypeError, error % (type(i), obj_type)
            # obj is an empty list
            except IndexError:
                self._obj = []
	    # All the objects in the list are of the same type
	    else:
		self._obj = list(obj)
        else:
            log.warn('Modifier.obj__set(): obj must be list or tuple')
	    self._obj = []

    def prop__get(self):
        """The property of the object class"""
        return self._prop

    def prop__set(self, prop):
        # Check if prop is in the namespace of the obj class
        try:
            attrib = self._obj[0].__getattribute__(prop)
            self._prop = prop
            # Store a pointer to the adapted lerp func
            if isinstance(attrib, (list, tuple)):
                self._lerp = maths.lerp_seq
                if self._operation == RELATIVE:
                    log.warn('Modifier.prop__set(): RELATIVE operation are not yet supported for tuple/list property type')
                    self._operation = ABSOLUTE
            else:
                self._lerp = maths.lerp
            
            if self._operation == RELATIVE:
                self._save_property_initial_value()
            
        except TypeError:
            log.err('Modifier.prop__set(): prop must be string')
            self._prop = None
        except IndexError:
            log.err('Modifier.prop__set(): obj list is empty')
            self._prop = None
        except AttributeError:
            log.err('Modifier.prop__set(): No attribute %s in %r',
                     prop, self._obj[0])
            self._prop = None

    def timeline__get(self):
        """The timeline (list of keyframes) to interpolate"""
        return self._timeline

    def timeline__set(self, timeline):
	if isinstance(timeline, (list, tuple)):
            try:
                for i in timeline:
                    if not isinstance(i, keyframe.KeyFrame):
                        raise TypeError, 'Not a keyframe.KeyFrame object'
            # obj is an empty list
            except IndexError:
                self._timeline = []
	    # All the objects in the list are KeyFrame
            # FIXME: A sort on the KeyFrame.time parameter is needed
            #        here because we can't assume the user give us a
            #        sorted list.
	    else:
		self._timeline = list(timeline)
        else:
            log.warn('Modifier.timeline__set(): ' \
		     'timeline must be list or tuple')
	    self._timeline = []
        self._timeline_len = len(self._timeline)
        if self._timeline_len < 2:
            raise TypeError, 'timeline must at least contain 2 KeyFrame'

    def interpolation__get(self):
        """The interpolation type"""
        return self._interpolation

    def interpolation__set(self, interpolation):
        if interpolation == LINEAR or interpolation == SPLINE:
            self._interpolation = interpolation
        else:
            log.warn('Modifier.interpolation__set():' \
                     'Bad interpolation type, reseting to LINEAR.')
            self._interpolation = LINEAR

    def _save_property_initial_value(self): 
        """
        """
        for obj in self._obj:
            value = obj.__getattribute__(self._prop)
            #print "for %s, save property %s with value %s"%(obj,self._prop, value)
            self._prop_initial_values.append( value )
        
    # Public methods

    def update(self, factor):
	"""
	"""
        # Search the two KeyFrame to interpolate. At the end i-1 is
        # the first KeyFrame index, and i the second.
        # We assume here that self._timeline is sorted.
        i = 0
        while self._timeline[i].time < factor:
            i += 1
            # The user give us a partial timeline finishing before 1.0.
            if i >= self._timeline_len:
                return

        # Expand the factor from the time range
        # [self._timeline[i-1].time, self._timeline[i].time]
        # to the range [0, 1].
        new_factor = (factor - self._timeline[i-1].time) \
                      / (self._timeline[i].time - \
                         self._timeline[i-1].time)
        # Linearly interpolate the new_factor between the KeyFrame values.
        value = self._lerp(self._timeline[i-1].value,
                           self._timeline[i].value, new_factor)

        # Let's modify the objects.
        rank = 0
        for obj in self._obj:
            if self._operation == RELATIVE:
                initial_value = self._prop_initial_values[rank]
                obj.__setattr__(self._prop, initial_value + value)
            else:
                obj.__setattr__(self._prop, value)
            rank += 1
