# -*- mode: python; coding: utf-8 -*-
# Pigment rendering library
#
# Copyright © 2006 Fluendo Embedded S.L. <www.fluendo.com>
# All rights reserved.
#
# This file is available under one of two license agreements.
#
# This file is licensed under the GPL version 2. See "LICENSE.GPL" in
# the root of this distribution including a special exception to use
# Pigment with Fluendo's plugins.
#
# The GPL part of Pigment is also available under a commercial licensing
# agreement from Fluendo. See "LICENSE.Pigment" in the root directory
# of this distribution package for details on that license.

"""
The widget base module.
"""

import pgmrender
from pgm import log
from pgm.util import classinit, maths
from pgm.message import signal

# Ordering constants
FAR, Z_ORDERED, NEAR = 0, 1, 2

order = {FAR : pgmrender.DRAWABLE_FAR,
         Z_ORDERED : pgmrender.DRAWABLE_Z_ORDERED,
         NEAR : pgmrender.DRAWABLE_NEAR}

BOTTOM_RIGHT = 0
BOTTOM_LEFT =1 
UP_LEFT = 2
UP_RIGHT = 3
CENTER = 4
UP_CENTER = 5
BOTTOM_CENTER = 6
LEFT_CENTER = 7
RIGHT_CENTER = 8

class Widget(object):
    """Widget base class
    """

    # Allows property fget/fset/fdel/doc overriding
    __metaclass__ = classinit.ClassInitMeta
    __classinit__ = classinit.build_properties

    # Cursor entered widgets in the current and last event handling pass.
    _entered = []
    _last_entered = []

    sig_key_press_event = signal.Signal('key-press-event', (object,))


    def __init__(self, size, position):
        self.size_mode = BOTTOM_RIGHT
        self._width = 0
        self._height = 0
        self._window = None
        self._visible = True
        self.size = size
        self.position = position
        self.translation = (0,0,0)
        self.rotation = (0,0,0)
        self._ordering = Z_ORDERED
        self._opacity = 255
        self._opacity_offset = 0
        self._parent = None
        #self._had_pointer = False

        # Widget signal registrations
        self.sig_button_press_event = signal.Signal('button-press-event',
                                                    (object,))
        self.sig_button_release_event = signal.Signal('button-release-event',
                                                      (object,))
        self.sig_delete_event = signal.Signal('delete-event')
        self.sig_enter_notify_event = signal.Signal('enter-notify-event')
        self.sig_leave_notify_event = signal.Signal('leave-notify-event')
        self.sig_key_release_event = signal.Signal('key-release-event',
                                                   (object,))
        self.sig_motion_notify_event = signal.Signal('motion-notify-event',
                                                     (object,))
        self.sig_scroll_event = signal.Signal('scroll-event')
        self.sig_state_changed = signal.Signal('state-changed', (object,))

    # Property definitions

    def width__get(self):
        """The width size"""
        return self._width
    def width__set(self, width):
        self._width = max(0, width)

    def height__get(self):
        """The height size"""
        return self._height
    def height__set(self, height):
        self._height = max(0, height) 
    
    def size_mode__get(self):
        """The size mode. can be : BOTTOM_RIGHT ,BOTTOM_LEFT, UP_LEFT, UP_RIGHT"""
        return self._size_mode
    def size_mode__set(self, size_mode):
        self._size_mode =  min(8, max(0,size_mode) )

    def size__get(self):
        """The size"""
        return (self._width, self._height)
        
    def size__set(self, size):
        self._width, self._height = max(0, size[0]), max(0, size[1])
        
    def x__get(self):
        """The x position"""
        return self._x
    def x__set(self, x):
        self._x = x

    def y__get(self):
        """The y position"""
        return self._y
    def y__set(self, y):
        self._y = y

    def position__get(self):
        """The position"""
        return (self._x, self._y)
    def position__set(self, position):
        self._x, self._y = position

    def tx__get(self):
        """The x translate position"""
        return self._tx
    def tx__set(self, tx):
        self._tx = tx

    def ty__get(self):
        """The y translate position"""
        return self._ty
    def ty__set(self, ty):
        self._ty = ty

    def tz__get(self):
        """The z translate position"""
        return self._tz
    def tz__set(self, tz):
        self._tz = tz

    def translation__get(self):
        """The translate position"""
        return (self._tx, self._ty, self._tz)
    def translation__set(self, translation):
        self._tx, self._ty, self._tz = translation

    def rx__get(self):
        """The x rotate position"""
        return self._rx
    def rx__set(self, rx):
        self._rx = rx

    def ry__get(self):
        """The y rotate position"""
        return self._ry
    def ry__set(self, ry):
        self._ry = ry

    def rz__get(self):
        """The z rotate position"""
        return self._rz
    def rz__set(self, rz):
        self._rz = rz

    def rotation__get(self):
        """The rotate position"""
        return (self._rx, self._ry, self._rz)
    def rotation__set(self, rotation):
        self._rx, self._ry, self._rz = rotation

    def visible__get(self):
        """The visibility"""
        return self._visible
    def visible__set(self, visible):
        if visible == True or visible == False:
            self._visible = visible
        else:
            self._visible = True

    def opacity__get(self):
        """The opacity"""
        return maths.clamp(self._opacity + self._opacity_offset, 0, 255)
    def opacity__set(self, opacity):
        self._opacity = maths.clamp(opacity, 0, 255)
        
    def opacity_offset__get(self):
        """The opacity offset"""
        return self._opacity_offset
    def opacity_offset__set(self, opacity_offset):
        self._opacity_offset = opacity_offset

    def ordering__get(self):
        """The ordering style"""
        return self._ordering
    def ordering__set(self, ordering):
        if ordering == NEAR or ordering == Z_ORDERED or ordering == FAR:
            self._ordering = ordering
        else:
            log.warn('Widget.ordering__set(): Bad ordering parameter')

    def parent__get(self):
        """The parent Widget"""
        return self._parent

    def window__get(self):
        """The binded window"""
        return self._window

    # Public methods

    def is_ancestor(self, widget):
        """
        """
        pass

    def ponderate_opacity(self, ponderation):
        """
        """
        self.opacity *= ponderation

    def get_signal_name(self, event):
        """
        """
        event_map = {
            pgmrender.KEY_PRESS: 'sig_key_press_event',
            pgmrender.KEY_RELEASE: 'sig_key_release_event',
            pgmrender.BUTTON_PRESS: 'sig_button_press_event',
            pgmrender.BUTTON_RELEASE: 'sig_button_release_event', 
        }
        return event_map.get(event.type)
    
    # Private methods

    def _set_window(self, window):
        """
        """
        self._window = window

    def _has_pointer(self, event):
        """
        """
        pass
        #return self._position[0] <= event.position[0] and \
        #self._position[0] + self._size[0] >= event.position[0] and \
        #self._position[1] <= event.position[1] and \
        #self._position[1] + self._size[1] >= event.position[1]

    @staticmethod
    def _get_renderer(self):
        """
        FIXME
        """
        pass
